// This file was adapted to usage in non-GNU environments
// (especially under wxWindows) by Stefan Hammes (steve),
// (stefan.hammes@urz.uni-heidelberg.de)
//
// There were also some minor extensions made. They can
// be found at the end of the 'wxString' constructor.
//
// To use this code indepently from libg++ all classes
// were renamed by adding a 'wx' in front of their names!
//
// 01.12.94
//
// NOTE: the DEC C++ compiler doesn't allow friend members to be
//       declared NOT inline in the class and inline afterwards.
//       therefore add the inline qualifier to the declarations
//       in the class.
// NOTE: the DEC C++ compiler doesn't allow array parameters of the
//       class in definition. therefore the split/join members are
//       not available under VMS
// Stefan Hammes (steve), 03.12.94

#ifdef wx_msw
#define VMSinline /**/ 
#else
#define VMSinline inline
#endif

// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of the GNU C++ Library.  This library is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This library is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#ifndef _wxString_h
#ifdef __GNUG__
#pragma interface
#endif
#define _wxString_h 1

#include <iostream.h>
#include <stdio.h> // for readline

#if defined(wx_msw) || defined(wx_motif)
#include "wx_obj.h"
#endif

#ifdef VMS
// this is included in [.contrib.wxstring]wxstring.h under VMS,
// but in \include\base\wxstring.h under DOS or UNIX
#include "wxregex.h"
#else
// must be done separately in wxstring.cc before including wxstring.h #include "wxregex.h"
#endif

struct wxStrRep                     // internal wxString representations
{
  unsigned short    len;         // string length 
  unsigned short    sz;          // allocated space
  char              s[1];        // the string starts here 
                                 // (at least 1 char for trailing null)
                                 // allocated & expanded via non-public fcts
};

// primitive ops on StrReps -- nearly all wxString fns go through these.

wxStrRep*     Salloc(wxStrRep*, const char*, int, int);
wxStrRep*     Scopy(wxStrRep*, const wxStrRep*);
wxStrRep*     Sresize(wxStrRep*, int);
wxStrRep*     Scat(wxStrRep*, const char*, int, const char*, int);
wxStrRep*     Scat(wxStrRep*, const char*, int,const char*,int, const char*,int);
wxStrRep*     Sprepend(wxStrRep*, const char*, int);
wxStrRep*     Sreverse(const wxStrRep*, wxStrRep*);
wxStrRep*     Supcase(const wxStrRep*, wxStrRep*);
wxStrRep*     Sdowncase(const wxStrRep*, wxStrRep*);
wxStrRep*     Scapitalize(const wxStrRep*, wxStrRep*);

// These classes need to be defined in the order given

class wxString;
class wxSubString;           

class wxSubString            
#if defined(wx_msw) || defined(wx_motif)
  : public wxObject
#endif
{
  friend class      wxString;
protected:

  wxString&           S;        // The wxString I'm a substring of
  unsigned short    pos;      // starting position in S's rep
  unsigned short    len;      // length of substring

  void              assign(const wxStrRep*, const char*, int = -1);
                    wxSubString(wxString& x, int p, int l);
                    wxSubString(const wxSubString& x);

public:

// Note there are no public constructors. SubStrings are always
// created via wxString operations

                   ~wxSubString();

  wxSubString&        operator =  (const wxString&     y);
  wxSubString&        operator =  (const wxSubString&  y);
  wxSubString&        operator =  (const char* t);
  wxSubString&        operator =  (char        c);

// return 1 if target appears anywhere in wxSubString; else 0

  int               contains(char        c) const;
  int               contains(const wxString&     y) const;
  int               contains(const wxSubString&  y) const;
  int               contains(const char* t) const;
  int               contains(const wxRegex&       r) const;

// return 1 if target matches entire wxSubString

  int               matches(const wxRegex&  r) const;

// IO 

  friend ostream&   operator<<(ostream& s, const wxSubString& x);

// status

  unsigned int      length() const;
  int               empty() const;
  const char*       chars() const;

  int               OK() const; 

};

class wxString
#if defined(wx_msw) || defined(wx_motif)
  : public wxObject
#endif
{
  friend class      wxSubString;

protected:
  wxStrRep*           rep;   // Strings are pointers to their representations

// some helper functions

  int               search(int, int, const char*, int = -1) const;
  int               search(int, int, char) const;
  int               match(int, int, int, const char*, int = -1) const;
  int               _gsub(const char*, int, const char* ,int);
  int               _gsub(const wxRegex&, const char*, int);
  wxSubString         _substr(int, int);

public:

// constructors & assignment

                    wxString();
                    wxString(const wxString& x);
                    wxString(const wxSubString&  x);
                    wxString(const char* t);
                    wxString(const char* t, int len);
                    wxString(char c);

                    ~wxString();

  wxString&           operator =  (const wxString&     y);
  wxString&           operator =  (const char* y);
  wxString&           operator =  (char        c);
  wxString&           operator =  (const wxSubString&  y);

// concatenation

  wxString&           operator += (const wxString&     y); 
  wxString&           operator += (const wxSubString&  y);
  wxString&           operator += (const char* t);
  wxString&           operator += (char        c);

  void              prepend(const wxString&     y); 
  void              prepend(const wxSubString&  y);
  void              prepend(const char* t);
  void              prepend(char        c);


// procedural versions:
// concatenate first 2 args, store result in last arg

  VMSinline friend void     cat(const wxString&, const wxString&, wxString&);
  VMSinline friend void     cat(const wxString&, const wxSubString&, wxString&);
  VMSinline friend void     cat(const wxString&, const char*, wxString&);
  VMSinline friend void     cat(const wxString&, char, wxString&);

  VMSinline friend void     cat(const wxSubString&, const wxString&, wxString&);
  VMSinline friend void     cat(const wxSubString&, const wxSubString&, wxString&);
  VMSinline friend void     cat(const wxSubString&, const char*, wxString&);
  VMSinline friend void     cat(const wxSubString&, char, wxString&);

  VMSinline friend void     cat(const char*, const wxString&, wxString&);
  VMSinline friend void     cat(const char*, const wxSubString&, wxString&);
  VMSinline friend void     cat(const char*, const char*, wxString&);
  VMSinline friend void     cat(const char*, char, wxString&);

// double concatenation, by request. (yes, there are too many versions, 
// but if one is supported, then the others should be too...)
// Concatenate first 3 args, store in last arg

  VMSinline friend void     cat(const wxString&,const wxString&, const wxString&,wxString&);
  VMSinline friend void     cat(const wxString&,const wxString&,const wxSubString&,wxString&);
  VMSinline friend void     cat(const wxString&,const wxString&, const char*, wxString&);
  VMSinline friend void     cat(const wxString&,const wxString&, char, wxString&);
  VMSinline friend void     cat(const wxString&,const wxSubString&,const wxString&,wxString&);
  VMSinline friend void     cat(const wxString&,const wxSubString&,const wxSubString&,wxString&);
  VMSinline friend void     cat(const wxString&,const wxSubString&, const char*, wxString&);
  VMSinline friend void     cat(const wxString&,const wxSubString&, char, wxString&);
  VMSinline friend void     cat(const wxString&,const char*, const wxString&,    wxString&);
  VMSinline friend void     cat(const wxString&,const char*, const wxSubString&, wxString&);
  VMSinline friend void     cat(const wxString&,const char*, const char*, wxString&);
  VMSinline friend void     cat(const wxString&,const char*, char, wxString&);

  VMSinline friend void     cat(const char*, const wxString&, const wxString&,wxString&);
  VMSinline friend void     cat(const char*,const wxString&,const wxSubString&,wxString&);
  VMSinline friend void     cat(const char*,const wxString&, const char*, wxString&);
  VMSinline friend void     cat(const char*,const wxString&, char, wxString&);
  VMSinline friend void     cat(const char*,const wxSubString&,const wxString&,wxString&);
  VMSinline friend void     cat(const char*,const wxSubString&,const wxSubString&,wxString&);
  VMSinline friend void     cat(const char*,const wxSubString&, const char*, wxString&);
  VMSinline friend void     cat(const char*,const wxSubString&, char, wxString&);
  VMSinline friend void     cat(const char*,const char*, const wxString&,    wxString&);
  VMSinline friend void     cat(const char*,const char*, const wxSubString&, wxString&);
  VMSinline friend void     cat(const char*,const char*, const char*, wxString&);
  VMSinline friend void     cat(const char*,const char*, char, wxString&);


// searching & matching

// return position of target in string or -1 for failure

  int               index(char        c, int startpos = 0) const;      
  int               index(const wxString&     y, int startpos = 0) const;      
  int               index(const wxSubString&  y, int startpos = 0) const;      
  int               index(const char* t, int startpos = 0) const;  
  int               index(const wxRegex&      r, int startpos = 0) const;       

// return 1 if target appears anyhere in wxString; else 0

  int               contains(char        c) const;
  int               contains(const wxString&     y) const;
  int               contains(const wxSubString&  y) const;
  int               contains(const char* t) const;
  int               contains(const wxRegex&      r) const;

// return 1 if target appears anywhere after position pos 
// (or before, if pos is negative) in wxString; else 0

  int               contains(char        c, int pos) const;
  int               contains(const wxString&     y, int pos) const;
  int               contains(const wxSubString&  y, int pos) const;
  int               contains(const char* t, int pos) const;
  int               contains(const wxRegex&      r, int pos) const;

// return 1 if target appears at position pos in wxString; else 0

  int               matches(char        c, int pos = 0) const;
  int               matches(const wxString&     y, int pos = 0) const;
  int               matches(const wxSubString&  y, int pos = 0) const;
  int               matches(const char* t, int pos = 0) const;
  int               matches(const wxRegex&      r, int pos = 0) const;

//  return number of occurences of target in wxString

  int               freq(char        c) const; 
  int               freq(const wxString&     y) const;
  int               freq(const wxSubString&  y) const;
  int               freq(const char* t) const;

// wxSubString extraction

// Note that you can't take a substring of a const wxString, since
// this leaves open the possiblility of indirectly modifying the
// wxString through the wxSubString

  wxSubString         at(int         pos, int len);
  wxSubString         operator () (int         pos, int len); // synonym for at

  wxSubString         at(const wxString&     x, int startpos = 0); 
  wxSubString         at(const wxSubString&  x, int startpos = 0); 
  wxSubString         at(const char* t, int startpos = 0);
  wxSubString         at(char        c, int startpos = 0);
  wxSubString         at(const wxRegex&      r, int startpos = 0); 

  wxSubString         before(int          pos);
  wxSubString         before(const wxString&      x, int startpos = 0);
  wxSubString         before(const wxSubString&   x, int startpos = 0);
  wxSubString         before(const char*  t, int startpos = 0);
  wxSubString         before(char         c, int startpos = 0);
  wxSubString         before(const wxRegex&       r, int startpos = 0);

  wxSubString         through(int          pos);
  wxSubString         through(const wxString&      x, int startpos = 0);
  wxSubString         through(const wxSubString&   x, int startpos = 0);
  wxSubString         through(const char*  t, int startpos = 0);
  wxSubString         through(char         c, int startpos = 0);
  wxSubString         through(const wxRegex&       r, int startpos = 0);

  wxSubString         from(int          pos);
  wxSubString         from(const wxString&      x, int startpos = 0);
  wxSubString         from(const wxSubString&   x, int startpos = 0);
  wxSubString         from(const char*  t, int startpos = 0);
  wxSubString         from(char         c, int startpos = 0);
  wxSubString         from(const wxRegex&       r, int startpos = 0);

  wxSubString         after(int         pos);
  wxSubString         after(const wxString&     x, int startpos = 0);
  wxSubString         after(const wxSubString&  x, int startpos = 0);
  wxSubString         after(const char* t, int startpos = 0);
  wxSubString         after(char        c, int startpos = 0);
  wxSubString         after(const wxRegex&      r, int startpos = 0);


// deletion

// delete len chars starting at pos
  void              del(int         pos, int len);

// delete the first occurrence of target after startpos

  void              del(const wxString&     y, int startpos = 0);
  void              del(const wxSubString&  y, int startpos = 0);
  void              del(const char* t, int startpos = 0);
  void              del(char        c, int startpos = 0);
  void              del(const wxRegex&      r, int startpos = 0);

// global substitution: substitute all occurrences of pat with repl

  int               gsub(const wxString&     pat, const wxString&     repl);
  int               gsub(const wxSubString&  pat, const wxString&     repl);
  int               gsub(const char* pat, const wxString&     repl);
  int               gsub(const char* pat, const char* repl);
  int               gsub(const wxRegex&      pat, const wxString&     repl);

// friends & utilities

// split string into array res at separators; return number of elements
#ifndef VMS
  // VMS doesn't allow the arrays here, (steve) 03.12.94
  friend int        split(const wxString& x, wxString res[], int maxn, 
                          const wxString& sep);
  friend int        split(const wxString& x, wxString res[], int maxn, 
                          const wxRegex&  sep);
  friend wxString     join(wxString src[], int n, const wxString& sep);
#endif

  friend wxString     common_prefix(const wxString& x, const wxString& y, 
                                  int startpos = 0);
  friend wxString     common_suffix(const wxString& x, const wxString& y, 
                                  int startpos = -1);
  friend wxString     replicate(char        c, int n);
  friend wxString     replicate(const wxString&     y, int n);


// simple builtin transformations

  VMSinline friend wxString     reverse(const wxString& x);
  VMSinline friend wxString     upcase(const wxString& x);
  VMSinline friend wxString     downcase(const wxString& x);
  VMSinline friend wxString     capitalize(const wxString& x);

// in-place versions of above

  void              reverse();
  void              upcase();
  void              downcase();
  void              capitalize();

// element extraction

  char&             operator [] (int i);
  char              elem(int i) const;
  char              firstchar() const;
  char              lastchar() const;

// conversion

                    operator const char*() const;
  const char*       chars() const;


// IO

  VMSinline friend ostream&   operator<<(ostream& s, const wxString& x);
  friend ostream&   operator<<(ostream& s, const wxSubString& x);
  friend istream&   operator>>(istream& s, wxString& x);

  friend int        readline(istream& s, wxString& x, 
                             char terminator = '\n',
                             int discard_terminator = 1);
// status

  unsigned int      length() const;
  int               empty() const;

// preallocate some space for wxString
  void              alloc(int newsize);

// report current allocation (not length!)

  int               allocation() const;


  void     error(const char* msg) const;

  int               OK() const;
  
//********************************************************************************
//********************************************************************************
//********************************************************************************
// additions by steve

// formatted assignment
  void sprintf(const char *fmt, ...);
// we do not use the 'sprintf' constructor anymore, because with that
// constructor, every initialisation with a string would go through
// sprintf and this is not desireable, because sprintf interprets
// some characters. with the above function we can write i.e.:
//
// wxString msg; msg.sprintf("Processing item %d\n",count);
  
// strip chars at the front and/or end
// stripType is defined for bitwise ORing!
  enum              stripType {leading = 0x1, trailing = 0x2, both = 0x3};
  wxSubString       strip(stripType s=trailing, char c=' ');

  // the funny casting is necessary to distinguish between the 'del' members  
  void              removelast() { del((int)(length()-1),(int)1); } 

  friend int        readline(FILE *f, wxString& x, 
                             char terminator = '\n',
                             int discard_terminator = 1);

// classification (should be capital, because of ctype.h macros!)
  int               isAscii() const;
  int               isWord() const;
  int               isNumber() const;
  int               isNull() const;
  int               isDefined() const { return(! isNull()); }

// here is a very, very, very ugly macro, but it makes things more
// transparent in cases, where a library function requires a 
// (char *) argument. this is especially the case in wxWindows,
// where all char-arguments are (char *) and not (const char *).
// this macro should only be used in such cases and NOT to
// modify the internal data.
// the conventional way would be 'function((char *)string.chars())'.
// with the macro this can be achieved by 'function(wxCHARARG(string))'.
// this makes it better clear, that the usage should be confined
// to arguments!

#define wxCHARARG(s) ((char *)(s).chars())  

//********************************************************************************
//********************************************************************************
//********************************************************************************
};

typedef wxString StrTmp; // for backward compatibility

// other externs

int        compare(const wxString&    x, const wxString&     y);
int        compare(const wxString&    x, const wxSubString&  y);
int        compare(const wxString&    x, const char* y);
int        compare(const wxSubString& x, const wxString&     y);
int        compare(const wxSubString& x, const wxSubString&  y);
int        compare(const wxSubString& x, const char* y);
int        fcompare(const wxString&   x, const wxString&     y); // ignore case

extern wxStrRep  _nilStrRep;
extern wxString _nilString;

// other inlines

VMSinline wxString operator + (const wxString& x, const wxString& y);
VMSinline wxString operator + (const wxString& x, const wxSubString& y);
VMSinline wxString operator + (const wxString& x, const char* y);
VMSinline wxString operator + (const wxString& x, char y);
VMSinline wxString operator + (const wxSubString& x, const wxString& y);
VMSinline wxString operator + (const wxSubString& x, const wxSubString& y);
VMSinline wxString operator + (const wxSubString& x, const char* y);
VMSinline wxString operator + (const wxSubString& x, char y);
VMSinline wxString operator + (const char* x, const wxString& y);
VMSinline wxString operator + (const char* x, const wxSubString& y);

VMSinline int operator==(const wxString& x, const wxString& y); 
VMSinline int operator!=(const wxString& x, const wxString& y);
VMSinline int operator> (const wxString& x, const wxString& y);
VMSinline int operator>=(const wxString& x, const wxString& y);
VMSinline int operator< (const wxString& x, const wxString& y);
VMSinline int operator<=(const wxString& x, const wxString& y);
VMSinline int operator==(const wxString& x, const wxSubString&  y);
VMSinline int operator!=(const wxString& x, const wxSubString&  y);
VMSinline int operator> (const wxString& x, const wxSubString&  y);
VMSinline int operator>=(const wxString& x, const wxSubString&  y);
VMSinline int operator< (const wxString& x, const wxSubString&  y);
VMSinline int operator<=(const wxString& x, const wxSubString&  y);
VMSinline int operator==(const wxString& x, const char* t);
VMSinline int operator!=(const wxString& x, const char* t);
VMSinline int operator> (const wxString& x, const char* t);
VMSinline int operator>=(const wxString& x, const char* t);
VMSinline int operator< (const wxString& x, const char* t);
VMSinline int operator<=(const wxString& x, const char* t);
VMSinline int operator==(const wxSubString& x, const wxString& y);
VMSinline int operator!=(const wxSubString& x, const wxString& y);
VMSinline int operator> (const wxSubString& x, const wxString& y);
VMSinline int operator>=(const wxSubString& x, const wxString& y);
VMSinline int operator< (const wxSubString& x, const wxString& y);
VMSinline int operator<=(const wxSubString& x, const wxString& y);
VMSinline int operator==(const wxSubString& x, const wxSubString&  y);
VMSinline int operator!=(const wxSubString& x, const wxSubString&  y);
VMSinline int operator> (const wxSubString& x, const wxSubString&  y);
VMSinline int operator>=(const wxSubString& x, const wxSubString&  y);
VMSinline int operator< (const wxSubString& x, const wxSubString&  y);
VMSinline int operator<=(const wxSubString& x, const wxSubString&  y);
VMSinline int operator==(const wxSubString& x, const char* t);
VMSinline int operator!=(const wxSubString& x, const char* t);
VMSinline int operator> (const wxSubString& x, const char* t);
VMSinline int operator>=(const wxSubString& x, const char* t);
VMSinline int operator< (const wxSubString& x, const char* t);
VMSinline int operator<=(const wxSubString& x, const char* t);


// status reports, needed before defining other things

inline unsigned int wxString::length() const {  return rep->len; }
inline int         wxString::empty() const { return rep->len == 0; }
inline const char* wxString::chars() const { return &(rep->s[0]); }
inline int         wxString::allocation() const { return rep->sz; }
inline void        wxString::alloc(int newsize) { rep = Sresize(rep, newsize); }

inline unsigned int wxSubString::length() const { return len; }
inline int         wxSubString::empty() const { return len == 0; }
inline const char* wxSubString::chars() const { return &(S.rep->s[pos]); }


// constructors

inline wxString::wxString() 
  : rep(&_nilStrRep) {}
inline wxString::wxString(const wxString& x) 
  : rep(Scopy(0, x.rep)) {}
inline wxString::wxString(const char* t) 
  : rep(Salloc(0, t, -1, -1)) {}
inline wxString::wxString(const char* t, int tlen)
  : rep(Salloc(0, t, tlen, tlen)) {}
inline wxString::wxString(const wxSubString& y)
  : rep(Salloc(0, y.chars(), y.length(), y.length())) {}
inline wxString::wxString(char c) 
  : rep(Salloc(0, &c, 1, 1)) {}

inline wxString::~wxString() { if (rep != &_nilStrRep) delete rep; }

inline wxSubString::wxSubString(const wxSubString& x)
  :S(x.S), pos(x.pos), len(x.len) {}
inline wxSubString::wxSubString(wxString& x, int first, int l)
  :S(x), pos((unsigned short) first), len((unsigned short) l) {}

inline wxSubString::~wxSubString() {}

// assignment

inline wxString& wxString::operator =  (const wxString& y)
{ 
  rep = Scopy(rep, y.rep);
  return *this;
}

inline wxString& wxString::operator=(const char* t)
{
  rep = Salloc(rep, t, -1, -1);
  return *this;
}

inline wxString& wxString::operator=(const wxSubString&  y)
{
  rep = Salloc(rep, y.chars(), y.length(), y.length());
  return *this;
}

inline wxString& wxString::operator=(char c)
{
  rep = Salloc(rep, &c, 1, 1);
  return *this;
}


inline wxSubString& wxSubString::operator = (const char* ys)
{
  assign(0, ys);
  return *this;
}

inline wxSubString& wxSubString::operator = (char ch)
{
  assign(0, &ch, 1);
  return *this;
}

inline wxSubString& wxSubString::operator = (const wxString& y)
{
  assign(y.rep, y.chars(), y.length());
  return *this;
}

inline wxSubString& wxSubString::operator = (const wxSubString& y)
{
  assign(y.S.rep, y.chars(), y.length());
  return *this;
}

// Zillions of cats...

inline void cat(const wxString& x, const wxString& y, wxString& r)
{
  r.rep = Scat(r.rep, x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const wxString& x, const wxSubString& y, wxString& r)
{
  r.rep = Scat(r.rep, x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const wxString& x, const char* y, wxString& r)
{
  r.rep = Scat(r.rep, x.chars(), x.length(), y, -1);
}

inline void cat(const wxString& x, char y, wxString& r)
{
  r.rep = Scat(r.rep, x.chars(), x.length(), &y, 1);
}

inline void cat(const wxSubString& x, const wxString& y, wxString& r)
{
  r.rep = Scat(r.rep, x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const wxSubString& x, const wxSubString& y, wxString& r)
{
  r.rep = Scat(r.rep, x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const wxSubString& x, const char* y, wxString& r)
{
  r.rep = Scat(r.rep, x.chars(), x.length(), y, -1);
}

inline void cat(const wxSubString& x, char y, wxString& r)
{
  r.rep = Scat(r.rep, x.chars(), x.length(), &y, 1);
}

inline void cat(const char* x, const wxString& y, wxString& r)
{
  r.rep = Scat(r.rep, x, -1, y.chars(), y.length());
}

inline void cat(const char* x, const wxSubString& y, wxString& r)
{
  r.rep = Scat(r.rep, x, -1, y.chars(), y.length());
}

inline void cat(const char* x, const char* y, wxString& r)
{
  r.rep = Scat(r.rep, x, -1, y, -1);
}

inline void cat(const char* x, char y, wxString& r)
{
  r.rep = Scat(r.rep, x, -1, &y, 1);
}

inline void cat(const wxString& a, const wxString& x, const wxString& y, wxString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const wxString& a, const wxString& x, const wxSubString& y, wxString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const wxString& a, const wxString& x, const char* y, wxString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x.chars(), x.length(), y, -1);
}

inline void cat(const wxString& a, const wxString& x, char y, wxString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x.chars(), x.length(), &y, 1);
}

inline void cat(const wxString& a, const wxSubString& x, const wxString& y, wxString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const wxString& a, const wxSubString& x, const wxSubString& y, wxString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const wxString& a, const wxSubString& x, const char* y, wxString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x.chars(), x.length(), y, -1);
}

inline void cat(const wxString& a, const wxSubString& x, char y, wxString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x.chars(), x.length(), &y, 1);
}

inline void cat(const wxString& a, const char* x, const wxString& y, wxString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x, -1, y.chars(), y.length());
}

inline void cat(const wxString& a, const char* x, const wxSubString& y, wxString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x, -1, y.chars(), y.length());
}

inline void cat(const wxString& a, const char* x, const char* y, wxString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x, -1, y, -1);
}

inline void cat(const wxString& a, const char* x, char y, wxString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x, -1, &y, 1);
}


inline void cat(const char* a, const wxString& x, const wxString& y, wxString& r)
{
  r.rep = Scat(r.rep, a, -1, x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const char* a, const wxString& x, const wxSubString& y, wxString& r)
{
  r.rep = Scat(r.rep, a, -1, x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const char* a, const wxString& x, const char* y, wxString& r)
{
  r.rep = Scat(r.rep, a, -1, x.chars(), x.length(), y, -1);
}

inline void cat(const char* a, const wxString& x, char y, wxString& r)
{
  r.rep = Scat(r.rep, a, -1, x.chars(), x.length(), &y, 1);
}

inline void cat(const char* a, const wxSubString& x, const wxString& y, wxString& r)
{
  r.rep = Scat(r.rep, a, -1, x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const char* a, const wxSubString& x, const wxSubString& y, wxString& r)
{
  r.rep = Scat(r.rep, a, -1, x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const char* a, const wxSubString& x, const char* y, wxString& r)
{
  r.rep = Scat(r.rep, a, -1, x.chars(), x.length(), y, -1);
}

inline void cat(const char* a, const wxSubString& x, char y, wxString& r)
{
  r.rep = Scat(r.rep, a, -1, x.chars(), x.length(), &y, 1);
}

inline void cat(const char* a, const char* x, const wxString& y, wxString& r)
{
  r.rep = Scat(r.rep, a, -1, x, -1, y.chars(), y.length());
}

inline void cat(const char* a, const char* x, const wxSubString& y, wxString& r)
{
  r.rep = Scat(r.rep, a, -1, x, -1, y.chars(), y.length());
}

inline void cat(const char* a, const char* x, const char* y, wxString& r)
{
  r.rep = Scat(r.rep, a, -1, x, -1, y, -1);
}

inline void cat(const char* a, const char* x, char y, wxString& r)
{
  r.rep = Scat(r.rep, a, -1, x, -1, &y, 1);
}


// operator versions

inline wxString& wxString::operator +=(const wxString& y)
{
  cat(*this, y, *this);
  return *this;
}

inline wxString& wxString::operator +=(const wxSubString& y)
{
  cat(*this, y, *this);
  return *this;
}

inline wxString& wxString::operator += (const char* y)
{
  cat(*this, y, *this);
  return *this;
}

inline wxString& wxString:: operator +=(char y)
{
  cat(*this, y, *this);
  return *this;
}

// constructive concatenation

#if defined(__GNUG__) && !defined(_G_NO_NRV)

inline wxString operator + (const wxString& x, const wxString& y) return r;
{
  cat(x, y, r);
}

inline wxString operator + (const wxString& x, const wxSubString& y) return r;
{
  cat(x, y, r);
}

inline wxString operator + (const wxString& x, const char* y) return r;
{
  cat(x, y, r);
}

inline wxString operator + (const wxString& x, char y) return r;
{
  cat(x, y, r);
}

inline wxString operator + (const wxSubString& x, const wxString& y) return r;
{
  cat(x, y, r);
}

inline wxString operator + (const wxSubString& x, const wxSubString& y) return r;
{
  cat(x, y, r);
}

inline wxString operator + (const wxSubString& x, const char* y) return r;
{
  cat(x, y, r);
}

inline wxString operator + (const wxSubString& x, char y) return r;
{
  cat(x, y, r);
}

inline wxString operator + (const char* x, const wxString& y) return r;
{
  cat(x, y, r);
}

inline wxString operator + (const char* x, const wxSubString& y) return r;
{
  cat(x, y, r);
}

inline wxString reverse(const wxString& x) return r;
{
  r.rep = Sreverse(x.rep, r.rep);
}

inline wxString upcase(const wxString& x) return r;
{
  r.rep = Supcase(x.rep, r.rep);
}

inline wxString downcase(const wxString& x) return r;
{
  r.rep = Sdowncase(x.rep, r.rep);
}

inline wxString capitalize(const wxString& x) return r;
{
  r.rep = Scapitalize(x.rep, r.rep);
}

#else /* NO_NRV */

inline wxString operator + (const wxString& x, const wxString& y)
{
  wxString r;  cat(x, y, r);  return r;
}

inline wxString operator + (const wxString& x, const wxSubString& y) 
{
  wxString r; cat(x, y, r); return r;
}

inline wxString operator + (const wxString& x, const char* y) 
{
  wxString r; cat(x, y, r); return r;
}

inline wxString operator + (const wxString& x, char y) 
{
  wxString r; cat(x, y, r); return r;
}

inline wxString operator + (const wxSubString& x, const wxString& y) 
{
  wxString r; cat(x, y, r); return r;
}

inline wxString operator + (const wxSubString& x, const wxSubString& y) 
{
  wxString r; cat(x, y, r); return r;
}

inline wxString operator + (const wxSubString& x, const char* y) 
{
  wxString r; cat(x, y, r); return r;
}

inline wxString operator + (const wxSubString& x, char y) 
{
  wxString r; cat(x, y, r); return r;
}

inline wxString operator + (const char* x, const wxString& y) 
{
  wxString r; cat(x, y, r); return r;
}

inline wxString operator + (const char* x, const wxSubString& y) 
{
  wxString r; cat(x, y, r); return r;
}

inline wxString reverse(const wxString& x) 
{
  wxString r; r.rep = Sreverse(x.rep, r.rep); return r;
}

inline wxString upcase(const wxString& x) 
{
  wxString r; r.rep = Supcase(x.rep, r.rep); return r;
}

inline wxString downcase(const wxString& x) 
{
  wxString r; r.rep = Sdowncase(x.rep, r.rep); return r;
}

inline wxString capitalize(const wxString& x) 
{
  wxString r; r.rep = Scapitalize(x.rep, r.rep); return r;
}

#endif

// prepend

inline void wxString::prepend(const wxString& y)
{
  rep = Sprepend(rep, y.chars(), y.length());
}

inline void wxString::prepend(const char* y)
{
  rep = Sprepend(rep, y, -1); 
}

inline void wxString::prepend(char y)
{
  rep = Sprepend(rep, &y, 1); 
}

inline void wxString::prepend(const wxSubString& y)
{
  rep = Sprepend(rep, y.chars(), y.length());
}

// misc transformations


inline void wxString::reverse()
{
  rep = Sreverse(rep, rep);
}


inline void wxString::upcase()
{
  rep = Supcase(rep, rep);
}


inline void wxString::downcase()
{
  rep = Sdowncase(rep, rep);
}


inline void wxString::capitalize()
{
  rep = Scapitalize(rep, rep);
}

// element extraction

inline char&  wxString::operator [] (int i) 
{ 
  if (((unsigned)i) >= length()) error("invalid index");
  return rep->s[i];
}

inline char  wxString::elem (int i) const
{ 
  if (((unsigned)i) >= length()) error("invalid index");
  return rep->s[i];
}

inline char  wxString::firstchar() const
{ 
  return elem(0);
}

inline char  wxString::lastchar() const
{ 
  return elem(length() - 1);
}

// searching

inline int wxString::index(char c, int startpos) const
{
  return search(startpos, length(), c);
}

inline int wxString::index(const char* t, int startpos) const
{   
  return search(startpos, length(), t);
}

inline int wxString::index(const wxString& y, int startpos) const
{   
  return search(startpos, length(), y.chars(), y.length());
}

inline int wxString::index(const wxSubString& y, int startpos) const
{   
  return search(startpos, length(), y.chars(), y.length());
}

inline int wxString::index(const wxRegex& r, int startpos) const
{
  int unused;  return r.search(chars(), length(), unused, startpos);
}

inline int wxString::contains(char c) const
{
  return search(0, length(), c) >= 0;
}

inline int wxString::contains(const char* t) const
{   
  return search(0, length(), t) >= 0;
}

inline int wxString::contains(const wxString& y) const
{   
  return search(0, length(), y.chars(), y.length()) >= 0;
}

inline int wxString::contains(const wxSubString& y) const
{   
  return search(0, length(), y.chars(), y.length()) >= 0;
}

inline int wxString::contains(char c, int p) const
{
  return match(p, length(), 0, &c, 1) >= 0;
}

inline int wxString::contains(const char* t, int p) const
{
  return match(p, length(), 0, t) >= 0;
}

inline int wxString::contains(const wxString& y, int p) const
{
  return match(p, length(), 0, y.chars(), y.length()) >= 0;
}

inline int wxString::contains(const wxSubString& y, int p) const
{
  return match(p, length(), 0, y.chars(), y.length()) >= 0;
}

inline int wxString::contains(const wxRegex& r) const
{
  int unused;  return r.search(chars(), length(), unused, 0) >= 0;
}

inline int wxString::contains(const wxRegex& r, int p) const
{
  return r.match(chars(), length(), p) >= 0;
}


inline int wxString::matches(const wxSubString& y, int p) const
{
  return match(p, length(), 1, y.chars(), y.length()) >= 0;
}

inline int wxString::matches(const wxString& y, int p) const
{
  return match(p, length(), 1, y.chars(), y.length()) >= 0;
}

inline int wxString::matches(const char* t, int p) const
{
  return match(p, length(), 1, t) >= 0;
}

inline int wxString::matches(char c, int p) const
{
  return match(p, length(), 1, &c, 1) >= 0;
}

inline int wxString::matches(const wxRegex& r, int p) const
{
  int l = (p < 0)? -p : length() - p;
  return r.match(chars(), length(), p) == l;
}


inline int wxSubString::contains(const char* t) const
{   
  return S.search(pos, pos+len, t) >= 0;
}

inline int wxSubString::contains(const wxString& y) const
{   
  return S.search(pos, pos+len, y.chars(), y.length()) >= 0;
}

inline int wxSubString::contains(const wxSubString&  y) const
{   
  return S.search(pos, pos+len, y.chars(), y.length()) >= 0;
}

inline int wxSubString::contains(char c) const
{
  return S.search(pos, pos+len, c) >= 0;
}

inline int wxSubString::contains(const wxRegex& r) const
{
  int unused;  return r.search(chars(), len, unused, 0) >= 0;
}

inline int wxSubString::matches(const wxRegex& r) const
{
  return r.match(chars(), len, 0) == (int)len;
}


inline int wxString::gsub(const wxString& pat, const wxString& r)
{
  return _gsub(pat.chars(), pat.length(), r.chars(), r.length());
}

inline int wxString::gsub(const wxSubString&  pat, const wxString& r)
{
  return _gsub(pat.chars(), pat.length(), r.chars(), r.length());
}

inline int wxString::gsub(const wxRegex& pat, const wxString& r)
{
  return _gsub(pat, r.chars(), r.length());
}

inline int wxString::gsub(const char* pat, const wxString& r)
{
  return _gsub(pat, -1, r.chars(), r.length());
}

inline int wxString::gsub(const char* pat, const char* r)
{
  return _gsub(pat, -1, r, -1);
}



inline  ostream& operator<<(ostream& s, const wxString& x)
{
   s << x.chars(); return s;
}

// a zillion comparison operators

inline int operator==(const wxString& x, const wxString& y) 
{
  return compare(x, y) == 0; 
}

inline int operator!=(const wxString& x, const wxString& y)
{
  return compare(x, y) != 0; 
}

inline int operator>(const wxString& x, const wxString& y)
{
  return compare(x, y) > 0; 
}

inline int operator>=(const wxString& x, const wxString& y)
{
  return compare(x, y) >= 0; 
}

inline int operator<(const wxString& x, const wxString& y)
{
  return compare(x, y) < 0; 
}

inline int operator<=(const wxString& x, const wxString& y)
{
  return compare(x, y) <= 0; 
}

inline int operator==(const wxString& x, const wxSubString&  y) 
{
  return compare(x, y) == 0; 
}

inline int operator!=(const wxString& x, const wxSubString&  y)
{
  return compare(x, y) != 0; 
}

inline int operator>(const wxString& x, const wxSubString&  y)      
{
  return compare(x, y) > 0; 
}

inline int operator>=(const wxString& x, const wxSubString&  y)
{
  return compare(x, y) >= 0; 
}

inline int operator<(const wxString& x, const wxSubString&  y) 
{
  return compare(x, y) < 0; 
}

inline int operator<=(const wxString& x, const wxSubString&  y)
{
  return compare(x, y) <= 0; 
}

inline int operator==(const wxString& x, const char* t) 
{
  return compare(x, t) == 0; 
}

inline int operator!=(const wxString& x, const char* t) 
{
  return compare(x, t) != 0; 
}

inline int operator>(const wxString& x, const char* t)  
{
  return compare(x, t) > 0; 
}

inline int operator>=(const wxString& x, const char* t) 
{
  return compare(x, t) >= 0; 
}

inline int operator<(const wxString& x, const char* t)  
{
  return compare(x, t) < 0; 
}

inline int operator<=(const wxString& x, const char* t) 
{
  return compare(x, t) <= 0; 
}

inline int operator==(const wxSubString& x, const wxString& y) 
{
  return compare(y, x) == 0; 
}

inline int operator!=(const wxSubString& x, const wxString& y)
{
  return compare(y, x) != 0;
}

inline int operator>(const wxSubString& x, const wxString& y)      
{
  return compare(y, x) < 0;
}

inline int operator>=(const wxSubString& x, const wxString& y)     
{
  return compare(y, x) <= 0;
}

inline int operator<(const wxSubString& x, const wxString& y)      
{
  return compare(y, x) > 0;
}

inline int operator<=(const wxSubString& x, const wxString& y)     
{
  return compare(y, x) >= 0;
}

inline int operator==(const wxSubString& x, const wxSubString&  y) 
{
  return compare(x, y) == 0; 
}

inline int operator!=(const wxSubString& x, const wxSubString&  y)
{
  return compare(x, y) != 0;
}

inline int operator>(const wxSubString& x, const wxSubString&  y)      
{
  return compare(x, y) > 0;
}

inline int operator>=(const wxSubString& x, const wxSubString&  y)
{
  return compare(x, y) >= 0;
}

inline int operator<(const wxSubString& x, const wxSubString&  y) 
{
  return compare(x, y) < 0;
}

inline int operator<=(const wxSubString& x, const wxSubString&  y)
{
  return compare(x, y) <= 0;
}

inline int operator==(const wxSubString& x, const char* t) 
{
  return compare(x, t) == 0; 
}

inline int operator!=(const wxSubString& x, const char* t) 
{
  return compare(x, t) != 0;
}

inline int operator>(const wxSubString& x, const char* t)  
{
  return compare(x, t) > 0; 
}

inline int operator>=(const wxSubString& x, const char* t) 
{
  return compare(x, t) >= 0; 
}

inline int operator<(const wxSubString& x, const char* t)  
{
  return compare(x, t) < 0; 
}

inline int operator<=(const wxSubString& x, const char* t) 
{
  return compare(x, t) <= 0; 
}


// a helper needed by at, before, etc.

inline wxSubString wxString::_substr(int first, int l)
{
  if (first < 0 || (unsigned)(first + l) > length() )
    return wxSubString(_nilString, 0, 0) ;
  else 
    return wxSubString(*this, first, l);
}

#endif
